import easyvvuq as uq
from easyvvuq.actions import Actions, Encode, Decode, CreateRunDirectory, ExecuteQCGPJ, QCGPJPool
import chaospy as cp
import os

__license__ = "LGPL"


def test_qcgpj(tmpdir):
    params = {
        "temp_init": {
            "type": "float",
            "min": 0.0,
            "max": 100.0,
            "default": 95.0},
        "kappa": {
            "type": "float",
            "min": 0.0,
            "max": 0.1,
            "default": 0.025},
        "t_env": {
            "type": "float",
            "min": 0.0,
            "max": 40.0,
            "default": 15.0},
        "out_file": {
            "type": "string",
            "default": "output.csv"}}
    output_filename = params["out_file"]["default"]
    output_columns = ["te"]

    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cooling/cooling.template',
        delimiter='$',
        target_filename='cooling_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                output_columns=output_columns)

    vary = {
        "kappa": cp.Uniform(0.025, 0.075),
        "t_env": cp.Uniform(15, 25)
    }

    cooling_sampler = uq.sampling.PCESampler(vary=vary, polynomial_order=3)
    cooling_stats = uq.analysis.PCEAnalysis(sampler=cooling_sampler, qoi_cols=output_columns)

    cooling_action = uq.actions.ExecuteLocal(os.path.abspath("tests/cooling/cooling_model.py") +
                                         " cooling_in.json")

    qcgpj_action = ExecuteQCGPJ(cooling_action)
    actions = Actions(CreateRunDirectory('/tmp'), Encode(encoder), qcgpj_action, Decode(decoder))

    campaign = uq.Campaign(name='beam', params=params, actions=actions)
    campaign.set_sampler(cooling_sampler)

    with QCGPJPool() as qcgpj:
        campaign.execute(pool=qcgpj).collate()

    campaign.apply_analysis(cooling_stats)

if __name__ == "__main__":
    test_qcgpj("/tmp/")



import easyvvuq as uq
from easyvvuq.actions import Actions, Encode, Decode, CreateRunDirectory
import chaospy as cp
import os
import sys
import pytest

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# If cannonsim has not been built (to do so, run the Makefile in tests/cannonsim/src/)
# then skip this test
if not os.path.exists("tests/cannonsim/bin/cannonsim"):
    pytest.skip(
        "Skipping cannonsim test (cannonsim is not installed in tests/cannonsim/bin/)",
        allow_module_level=True)

CANNONSIM_PATH = os.path.realpath(os.path.expanduser("tests/cannonsim/bin/cannonsim"))


def test_multiencoder(tmpdir):

    # Set up a fresh campaign called "cannon"
    my_campaign = uq.Campaign(name='cannon', work_dir=tmpdir)

    # Define parameter space for the cannonsim app
    params = {
        "angle": {
            "type": "float",
            "min": 0.0,
            "max": 6.28,
            "default": 0.79},
        "air_resistance": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.2},
        "height": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 1.0},
        "time_step": {
            "type": "float",
            "min": 0.0001,
            "max": 1.0,
            "default": 0.01},
        "gravity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 9.8},
        "mass": {
            "type": "float",
            "min": 0.0001,
            "max": 1000.0,
            "default": 1.0},
        "velocity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 10.0}}

    # Specify a complicated directory hierarchy to test the DirectoryBuilder encoder
    directory_tree = {'dir1': {'dir2': {'dir3': None, 'dir4': None}}, 'dir5': {'dir6': None}}

    # Create a multiencoder combining a directory build, and two template encodes
    multiencoder = uq.encoders.MultiEncoder(

        uq.encoders.DirectoryBuilder(tree=directory_tree),

        uq.encoders.GenericEncoder(
            template_fname='tests/cannonsim/test_input/cannonsim.template',
            delimiter='#',
            target_filename='dir1/dir2/dir3/in.cannon'
        ),

        uq.encoders.GenericEncoder(
            template_fname='tests/cannonsim/test_input/cannonsim.template',
            delimiter='#',
            target_filename='dir5/dir6/in.cannon.2'
        )
    )

    # Create decoder and collater for the cannonsim app
    decoder = uq.decoders.SimpleCSV(
        target_filename='output.csv', output_columns=[
            'Dist', 'lastvx', 'lastvy'])
    actions = Actions(
        CreateRunDirectory('/tmp'),
        Encode(multiencoder),
        uq.actions.ExecuteLocal(
            os.path.abspath("tests/cannonsim/bin/cannonsim dir5/dir6/in.cannon.2") +
            " output.csv"),
        Decode(decoder))
    # Add the cannonsim app
    my_campaign.add_app(name="cannonsim",
                        params=params,
                        actions=actions)

    # Set the active app to be cannonsim (this is redundant when only one app
    # has been added)
    my_campaign.set_app("cannonsim")

    # Set up sampler
    sweep1 = {
        "angle": [0.1, 0.2, 0.3],
        "height": [2.0, 10.0],
        "velocity": [10.0, 10.1, 10.2]
    }
    sampler = uq.sampling.BasicSweep(sweep=sweep1)

    # Set the campaign to use this sampler
    my_campaign.set_sampler(sampler)

    reloaded_campaign = uq.Campaign('cannon', db_location=my_campaign.db_location)

    my_campaign.execute(sequential=True).collate()

    # Create a BasicStats analysis element and apply it to the campaign
    stats = uq.analysis.BasicStats(qoi_cols=['Dist', 'lastvx', 'lastvy'])
    my_campaign.apply_analysis(stats)


if __name__ == "__main__":
    test_multiencoder("/tmp/")



import pytest
import os
from unittest.mock import MagicMock
from easyvvuq.actions.execute_kubernetes import ExecuteKubernetes
import easyvvuq.actions.execute_kubernetes as execute_kubernetes

execute_kubernetes.config = MagicMock()
execute_kubernetes.V1ConfigMap = MagicMock()
execute_kubernetes.V1ObjectMeta = MagicMock()


def test_execute_kubernetes():
    kube_exec = ExecuteKubernetes(
        "orbitfold/easyvvuq:latest",
        "/EasyVVUQ/tutorials/sir /config/input.json && cat output.csv",
        output_file_name='output.csv')





"""This module provides an assortment of actions generally concerned with executing
simulations locally. Some Actions will also be useful when using Dask.
"""

import os
from pathlib import Path
import shutil
import subprocess
import dill
import copy

__license__ = "LGPL"


def local_execute(encoder, command, decoder, root='/tmp'):
    """A helper function for a simple local execution.
    It will create a directory under your specified root folder, encode the sampler output, execute a command
    and decode the results of the simulation.

    Parameters
    ----------
    encoder: Encoder
      an encoder to use
    command: list of str
      a command to run your simulation (same as argument to popen, e.g. ['ls', '-al'])
    decoder: Decoder
      a decoder to use
    root: str
      root folder, for example '/tmp' or if you want to use ram based filesystem it could be '/dev/shm'

    Returns
    -------
    Actions
    """
    return Actions(
        CreateRunDirectory(root),
        Encode(encoder),
        ExecuteLocal(command),
        Decode(decoder))


class CreateRunDirectory():
    """Creates a directory structure for storing simulation input and output files.

    Parameters
    ----------
    root: str
        Root directory to create a directory structure in.
    flatten: bool
        If set to True will result in a flat directory structure (each run gets a directory
        under root). If left as False will create a hierarchical structure. This is useful
        so as not to overload the filesystem.
    """

    def __init__(self, root, flatten=False):
        self.root = root
        self.flatten = flatten

    def start(self, previous=None):
        """Starts the action.

        Will read a `run_id` from a dictionary supplied by the previous `Action`.
        Will then create a directory structure based on the numerical value of the `run_id`.

        Returns
        -------
        dict
            A dictionary to be passed to the following `Action`.
        """
        run_id = previous['run_id']
        level1_a, level1_b = (int(run_id / 100 ** 4) * 100 ** 4,
                              int(run_id / 100 ** 4 + 1) * 100 ** 4)
        level2_a, level2_b = (int(run_id / 100 ** 3) * 100 ** 3,
                              int(run_id / 100 ** 3 + 1) * 100 ** 3)
        level3_a, level3_b = (int(run_id / 100 ** 2) * 100 ** 2,
                              int(run_id / 100 ** 2 + 1) * 100 ** 2)
        level4_a, level4_b = (int(run_id / 100 ** 1) * 100 ** 1,
                              int(run_id / 100 ** 1 + 1) * 100 ** 1)
        level1_dir = "runs_{}-{}/".format(level1_a, level1_b)
        level2_dir = "runs_{}-{}/".format(level2_a, level2_b)
        level3_dir = "runs_{}-{}/".format(level3_a, level3_b)
        level4_dir = "runs_{}-{}/".format(level4_a, level4_b)
        level5_dir = "run_{}".format(int(run_id))
        if self.flatten:
            path = os.path.join(self.root, previous['campaign_dir'], 'runs', level5_dir)
        else:
            path = os.path.join(self.root, previous['campaign_dir'], 'runs',
                                level1_dir, level2_dir, level3_dir, level4_dir, level5_dir)
        Path(path).mkdir(parents=True, exist_ok=True)
        previous['rundir'] = path
        self.result = previous
        return self.result

    def succeeded(self):
        """Has the `Action` finished successfully.

        Returns
        -------
        bool
            True if `Action` completed successfully. False otherwise.
        """
        return True


class Encode():
    def __init__(self, encoder):
        self.encoder = encoder

    def start(self, previous=None):
        self.encoder.encode(
            params=previous['run_info']['params'],
            target_dir=previous['rundir'])
        try:
            previous['encoder_filename'] = self.encoder.target_filename
        except AttributeError:
            pass
        return previous

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeeded(self):
        return True


class Decode():
    def __init__(self, decoder):
        self.decoder = decoder

    def start(self, previous=None):
        run_info = copy.copy(previous['run_info'])
        run_info['run_dir'] = previous['rundir']
        result = self.decoder.parse_sim_output(run_info)
        previous['result'] = result
        previous['decoder_filename'] = self.decoder.target_filename
        previous['collated'] = True
        return previous

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeded(self):
        return True


class CleanUp():
    def __init__(self):
        pass

    def start(self, previous=None):
        if not ('rundir' in previous.keys()):
            raise RuntimeError('must be used with actions that create a directory structure')
        shutil.rmtree(previous['rundir'])
        return previous

    def finished(self):
        return True

    def finalise(self):
        pass

    def succeeded(self):
        return True


class ExecutePython():
    def __init__(self, function):
        self.function = dill.dumps(function)
        self.params = None
        self.eval_result = None

    def start(self, previous=None):
        function = dill.loads(self.function)
        self.eval_result = function(previous['run_info']['params'])
        previous['result'] = self.eval_result
        previous['collated'] = True
        return previous

    def finished(self):
        if self.eval_result is None:
            return False
        else:
            return True

    def finalise(self):
        pass

    def succeeded(self):
        if not self.finished():
            raise RuntimeError('action did not finish yet')
        else:
            return True


class ExecuteLocal():
    def __init__(self, full_cmd, stdout=None, stderr=None):
        self.full_cmd = full_cmd.split()
        self.popen_object = None
        self.ret = None
        self._started = False
        self.stdout = stdout
        self.stderr = stderr

    def start(self, previous=None):
        target_dir = previous['rundir']
        if isinstance(self.stdout, str):
            stdout = open(os.path.join(target_dir, self.stdout), 'w')
        else:
            stdout = self.stdout
        if isinstance(self.stderr, str):
            stderr = open(os.path.join(target_dir, self.stderr), 'w')
        else:
            stderr = self.stderr
        self.ret = subprocess.run(
            self.full_cmd, cwd=target_dir,
            stdout=stdout, stderr=stderr)
        return previous

    def finished(self):
        return True

    def finalise(self):
        """Performs clean-up if necessary. In this case it isn't. I think.
        """
        pass

    def succeeded(self):
        """Will return True if the process finished successfully.
        It judges based on the return code and will return False
        if that code is not zero.
        """
        if self.ret != 0:
            return False
        else:
            return True


class Actions():
    def __init__(self, *args):
        self.actions = list(args)
        self.wrapper = lambda action, previous: action.start(previous)

    def set_wrapper(self, wrapper):
        """Adds a wrapper to be called on each Action.

        Parameters
        ----------
        wrapper: callable
            A function to call on each Action. Should pass through the return of the 
            start method.
        """
        self.wrapper = wrapper

    def start(self, previous=None):
        for action in self.actions:
            if not hasattr(action, 'start'):
                raise RuntimeError('action in the actions list does not provide a start method')
        previous = copy.copy(previous)
        run_id = previous['run_id']
        for action in self.actions:
            previous = self.wrapper(action, previous)
        self.result = previous
        assert(self.result['run_id'] == run_id)
        return previous

    def finished(self):
        return all([action.finished() for action in self.actions])

    def finalise(self):
        for action in self.actions:
            action.finalise()

    def succeeded(self):
        return all([action.succeeded() for action in self.actions])



"""A Decoder that can be used to get information from a YAML file.
Works identically to the JSON decoder. Look at the documentation of that
class for more information
"""

from easyvvuq.decoders.json import JSONDecoder
import logging
import yaml

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.Logger(__name__)


class YAMLDecoder(JSONDecoder):
    def _get_raw_data(self, out_path):
        """Reads in data from a YAML file.

        Parameters
        ----------
        out_path: str
            File name of a YAML file produced by the simulation code.
        """
        with open(out_path) as fd:
            return yaml.load(fd, yaml.SafeLoader)



import pytest
import time
from unittest.mock import MagicMock
from easyvvuq.actions import ActionPool


@pytest.fixture(scope="module", params=[False])
def action_pool():
    campaign = MagicMock()
    actions = MagicMock()
    inits = [MagicMock(), MagicMock(), MagicMock()]
    return ActionPool(campaign, actions, inits, sequential=False)


def test_action_pool_start(action_pool):
    action_pool.start()
    assert(action_pool.progress()['finished'] == 3)
    mock1 = MagicMock()
    mock1.running = MagicMock(return_value=True)
    mock1.done = MagicMock(return_value=False)
    mock1.result = MagicMock(return_value=False)
    mock2 = MagicMock()
    mock2.running = MagicMock(return_value=False)
    mock2.done = MagicMock(return_value=True)
    mock2.result = MagicMock(return_value=False)
    mock3 = MagicMock()
    mock3.running = MagicMock(return_value=False)
    mock3.done = MagicMock(return_value=True)
    mock3.result = MagicMock(return_value=True)
    mock4 = MagicMock()
    mock4.running = MagicMock(return_value=False)
    mock4.done = MagicMock(return_value=False)
    mock4.result = MagicMock(return_value=False)
    action_pool.futures = [mock1, mock2, mock3, mock4]
    progress = action_pool.progress()
    assert(progress['ready'] == 1)
    assert(progress['active'] == 1)
    assert(progress['finished'] == 1)
    assert(progress['failed'] == 1)



"""This module contains implementations of various Actions. Actions in
EasyVVUQ are responsible for anything that is related to the execution of
the simulation. That includes: actually executing the simulation, preparing
the input files, parsing the output files, creating directory structures
necessary to execute the simulation, cleaning up after, delegating work
to external execution back-ends such as Dask, etc.
"""

from .execute_local import ExecuteLocal, ExecutePython, CreateRunDirectory, Encode, Decode, local_execute
from .execute_local import CleanUp, Actions
from .execute_qcgpj import QCGPJPool, EasyVVUQBasicTemplate, ExecuteQCGPJ
from .execute_kubernetes import ExecuteKubernetes
from .execute_slurm import ExecuteSLURM
from .action_statuses import ActionPool

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"



"""A Decoder for CSV format files.
"""

import os
import logging
import csv
from easyvvuq import OutputType

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.Logger(__name__)


class SimpleCSV:
    """CSV Decoder.

    Parameters
    ----------
    target_filename: str
        Filename of a CSV file to decode.
    ouput_columns: list
        A list of column names that will be selected to appear in the output.
    """
    def __init__(self, target_filename, output_columns, dialect='excel'):
        if len(output_columns) == 0:
            msg = "output_columns cannot be empty."
            logger.error(msg)
            raise RuntimeError(msg)
        self.target_filename = target_filename
        self.output_columns = output_columns
        self.output_type = OutputType('sample')
        self.dialect = dialect

    @staticmethod
    def _get_output_path(run_info=None, outfile=None):
        """Constructs absolute path from the `target_filename` and the `run_dir` parameter
        in the `run_info` retrieved from the database.
        
        Parameters
        ----------
        run_info: dict
            Run info as retrieved from the database.
        outfile: str
            Filename of the file to be parsed.

        Returns
        -------
        str
            An absolute path to the output file in the run directory.
        """
        run_path = run_info['run_dir']
        if not os.path.isdir(run_path):
            raise RuntimeError(f"Run directory does not exist: {run_path}")
        return os.path.join(run_path, outfile)

    def parse_sim_output(self, run_info={}):
        """Parses the CSV file and converts it to the EasyVVUQ internal dictionary based
        format. The file is parsed in such a way that each column will appear as a vector
        QoI in the output dictionary.

        For example if the file contains the following data
        a,b
        1,2
        3,4

        And both `a` and `b` are specified as `output_columns` the output will look as follows
        {'a': [1, 3], 'b': [2, 4]}.

        Parameters
        ----------
        run_info: dict
            Information about the run (used to retrieve construct the absolute path
            to the CSV file that needs decoding.
        """
        out_path = self._get_output_path(run_info, self.target_filename)
        results = {}
        for column in self.output_columns:
            results[column] = []
        with open(out_path, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile, dialect=self.dialect)
            for row in reader:
                for column in self.output_columns:
                    try:
                        results[column].append(float(row[column]))
                    except ValueError:
                        results[column].append(row[column])
                    except KeyError:
                        raise RuntimeError('column not found in the csv file: {}'.format(column))
        return results



"""Provides class that allows access to an SQL Database that serves as the back-end to EasyVVUQ.


"""
import os
import json
import logging
import pandas as pd
import numpy as np
from sqlalchemy.sql import case
from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import text
from sqlalchemy.engine import Engine
from sqlalchemy import event
from .base import BaseCampaignDB
from easyvvuq import constants
from easyvvuq import ParamsSpecification
from easyvvuq.utils.helpers import easyvvuq_serialize, easyvvuq_deserialize


__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

COMMIT_RATE = 50000

logger = logging.getLogger(__name__)

Base = declarative_base()


class DBInfoTable(Base):
    """An SQLAlchemy schema for the database information table.
    """
    __tablename__ = 'db_info'
    id = Column(Integer, primary_key=True)
    next_run = Column(Integer)


class CampaignTable(Base):
    """An SQLAlchemy schema for the campaign information table.
    """
    __tablename__ = 'campaign_info'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    easyvvuq_version = Column(String)
    campaign_dir_prefix = Column(String)
    campaign_dir = Column(String)
    runs_dir = Column(String)
    sampler = Column(Integer, ForeignKey('sampler.id'))
    active_app = Column(Integer, ForeignKey('app.id'))


class AppTable(Base):
    """An SQLAlchemy schema for the app table.
    """
    __tablename__ = 'app'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    params = Column(String)
    actions = Column(String)


class RunTable(Base):
    """An SQLAlchemy schema for the run table.
    """
    __tablename__ = 'run'
    id = Column(Integer, primary_key=True)
    run_name = Column(String, index=True)
    app = Column(Integer, ForeignKey('app.id'))
    params = Column(String)
    status = Column(Integer)
    run_dir = Column(String)
    result = Column(String, default="{}")
    execution_info = Column(String, default="{}")
    campaign = Column(Integer, ForeignKey('campaign_info.id'))
    sampler = Column(Integer, ForeignKey('sampler.id'))
    iteration = Column(Integer, default=0)


class SamplerTable(Base):
    """An SQLAlchemy schema for the run table.
    """
    __tablename__ = 'sampler'
    id = Column(Integer, primary_key=True)
    sampler = Column(String)


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA synchronous = OFF")
    cursor.execute("PRAGMA journal_mode = OFF")
    cursor.close()


class CampaignDB(BaseCampaignDB):
    """An interface between the campaign database and the campaign.

    Parameters
    ----------
    location: str
       database URI as needed by SQLAlchemy
    """

    def __init__(self, location=None):
        if location is not None:
            self.engine = create_engine(location)
        else:
            self.engine = create_engine('sqlite://')
        self.commit_counter = 0
        session_maker = sessionmaker(bind=self.engine)
        self.session = session_maker()
        Base.metadata.create_all(self.engine, checkfirst=True)

    def resume_campaign(self, name):
        """Resumes campaign.

        Parameters
        ----------
        name: str
           Name of the Campaign to resume. Must already exist in the database.
        """
        info = self.session.query(
            CampaignTable).filter_by(name=name).first()
        if info is None:
            raise ValueError('Campaign with the given name not found.')
        db_info = self.session.query(DBInfoTable).first()
        self._next_run = db_info.next_run

    def create_campaign(self, info):
        """Creates a new campaign in the database.

        Parameters
        ----------
        info: CampaignInfo
            This `easyvvuq.data_structs.CampaignInfo` will contain information
            needed to construct the Campaign table.
        """
        is_db_empty = (self.session.query(CampaignTable).first() is None)
        version_check = self.session.query(
            CampaignTable).filter(CampaignTable.easyvvuq_version != info.easyvvuq_version).all()
        if (not is_db_empty) and (len(version_check) != 0):
            raise RuntimeError('Database contains campaign created with an incompatible' +
                               ' version of EasyVVUQ!')
        self._next_run = 1
        self.session.add(CampaignTable(**info.to_dict(flatten=True)))
        self.session.add(DBInfoTable(next_run=self._next_run))
        self.session.commit()

    def get_active_app(self):
        """Returns active app table.

        Returns
        -------
        AppTable
        """
        return self.session.query(AppTable, CampaignTable).filter(
            AppTable.id == CampaignTable.active_app).first()

    def campaign_exists(self, name):
        """Check if campaign specified by that name already exists.

        Parameters
        ----------
        name: str

        Returns
        -------
        bool
          True if such a campaign already exists, False otherwise
        """
        result = self.session.query(CampaignTable).filter(
            CampaignTable.name == name).all()
        return len(result) > 0

    def app(self, name=None):
        """Get app information. Specific applications selected by `name`,
        otherwise first entry in database 'app' selected.

        Parameters
        ----------
        name : str or None
            Name of selected app, if `None` given then first app will be
            selected.

        Returns
        -------
        dict
            Information about the application.
        """

        if name is None:
            selected = self.session.query(AppTable).all()
        else:
            selected = self.session.query(AppTable).filter_by(name=name).all()

        if len(selected) == 0:
            message = f'No entry for app: ({name}).'
            logger.critical(message)
            raise RuntimeError(message)

        selected_app = selected[0]

        app_dict = {
            'id': selected_app.id,
            'name': selected_app.name,
            'params': ParamsSpecification.deserialize(selected_app.params),
            'actions': selected_app.actions,
        }

        return app_dict

    def set_active_app(self, name):
        """Set an app specified by name as active.

        Parameters
        ----------
        name: str
           name of the app to set as active
        """
        selected = self.session.query(AppTable).filter_by(name=name).all()
        if len(selected) == 0:
            raise RuntimeError('no such app - {}'.format(name))
        assert(not (len(selected) > 1))
        app = selected[0]
        self.session.query(CampaignTable).update({'active_app': app.id})
        self.session.commit()

    def add_app(self, app_info):
        """Add application to the 'app' table.

        Parameters
        ----------
        app_info: AppInfo
            Application definition.
        """

        # Check that no app with same name exists
        name = app_info.name
        selected = self.session.query(AppTable).filter_by(name=name).all()
        if len(selected) > 0:
            message = (
                f'There is already an app in this database with name {name}'
                f'(found {len(selected)}).'
            )
            logger.critical(message)
            raise RuntimeError(message)

        app_dict = app_info.to_dict(flatten=True)

        db_entry = AppTable(**app_dict)
        self.session.add(db_entry)
        self.session.commit()

    def replace_actions(self, app_name, actions):
        """Replace actions for an app with a given name.

        Parameters
        ----------
        app_name: str
            Name of the app.
        actions: Actions
            `Actions` instance, will replace the current `Actions` of an app.
        """
        self.session.query(AppTable).filter_by(name=app_name).update(
            {'actions': easyvvuq_serialize(actions)})
        self.session.commit()

    def add_sampler(self, sampler_element):
        """Add new Sampler to the 'sampler' table.

        Parameters
        ----------
        sampler_element: Sampler
            An EasyVVUQ sampler.

        Returns
        -------
        int
            The sampler `id` in the database.
        """
        db_entry = SamplerTable(sampler=easyvvuq_serialize(sampler_element))

        self.session.add(db_entry)
        self.session.commit()

        return db_entry.id

    def update_sampler(self, sampler_id, sampler_element):
        """Update the state of the Sampler with id 'sampler_id' to
        that in the passed 'sampler_element'

        Parameters
        ----------
        sampler_id: int
            The id of the sampler in the db to update
        sampler_element: Sampler
            The sampler that should be used as the new state
        """

        selected = self.session.query(SamplerTable).get(sampler_id)
        selected.sampler = easyvvuq_serialize(sampler_element)
        self.session.commit()

    def resurrect_sampler(self, sampler_id):
        """Return the sampler object corresponding to id sampler_id in the database.
        It is deserialized from the state stored in the database.

        Parameters
        ----------
        sampler_id: int
            The id of the sampler to resurrect

        Returns
        -------
        Sampler
            The 'live' sampler object, deserialized from the state in the db
        """
        try:
            serialized_sampler = self.session.query(SamplerTable).get(sampler_id).sampler
            sampler = easyvvuq_deserialize(serialized_sampler.encode('utf-8'))
        except AttributeError:
            sampler = None
        return sampler

    def resurrect_app(self, app_name):
        """Return the 'live' encoder, decoder and collation objects corresponding to the app with
        name 'app_name' in the database. They are deserialized from the states previously
        stored in the database.

        Parameters
        ----------
        app_name: string
            Name of the app to resurrect

        Returns
        -------
        Actions
            The 'live' `Actions` object associated with this app. Used to execute the simulation
            associated with the app as well as do any pre- and post-processing.
        """
        app_info = self.app(app_name)
        actions = easyvvuq_deserialize(app_info['actions'])
        return actions

    def add_runs(self, run_info_list=None, run_prefix='run_', iteration=0):
        """Add list of runs to the `runs` table in the database.

        Parameters
        ----------
        run_info_list: List of RunInfo objects
            Each RunInfo object contains relevant run fields: params, status (where in the
            EasyVVUQ workflow is this RunTable), campaign (id number), sample, app
        run_prefix: str
            Prefix for run name
        iteration: int
            Iteration number used by iterative workflows. For example, MCMC. Can be left
            as default zero in other cases.
        """
        # Add all runs to RunTable
        commit_counter = 0
        for run_info in run_info_list:
            run_info.run_name = f"{run_prefix}{self._next_run}"
            run_info.iteration = iteration
            run = RunTable(**run_info.to_dict(flatten=True))
            self.session.add(run)
            self._next_run += 1
            commit_counter += 1
            if commit_counter % COMMIT_RATE == 0:
                self.session.commit()
        # Update run and ensemble counters in db
        db_info = self.session.query(DBInfoTable).first()
        db_info.next_run = self._next_run
        self.session.commit()

    @staticmethod
    def _run_to_dict(run_row):
        """Convert the provided row from 'runs' table into a dictionary

        Parameters
        ----------
        run_row: RunTable
            Information on a particular run in the database.

        Returns
        -------
        dict
            Contains run information (keys = run_name, params, status, sample,
            campaign and app)
        """

        run_info = {
            'run_name': run_row.run_name,
            'params': json.loads(run_row.params),
            'status': constants.Status(run_row.status),
            'sampler': run_row.sampler,
            'campaign': run_row.campaign,
            'app': run_row.app,
            'result': run_row.result,
            'run_dir': run_row.run_dir
        }

        return run_info

    def set_dir_for_run(self, run_name, run_dir, campaign=None, sampler=None):
        """Set the 'run_dir' path for the specified run in the database.

        Parameters
        ----------
        run_name: str
            Name of run to filter for.
        run_dir: str
            Directory path associated to set for this run.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sample id to filter for.
        """
        filter_options = {'run_name': run_name}
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler
        selected = self.session.query(RunTable).filter_by(**filter_options)
        if selected.count() != 1:
            logging.critical('Multiple runs selected - using the first')
        selected = selected.first()
        selected.run_dir = run_dir
        self.session.commit()

    def get_run_status(self, run_id, campaign=None, sampler=None):
        """Return the status (enum) for the run with name 'run_name' (and, optionally,
        filtering for campaign and sampler by id)

        Parameters
        ----------
        run_id: int
            id of the run
        campaign: int
            ID of the desired Campaign
        sampler: int
            ID of the desired Sampler

        Returns
        -------
        enum(Status)
            Status of the run.
        """
        filter_options = {'id': run_id}
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler
        selected = self.session.query(RunTable).filter_by(**filter_options)
        if selected.count() != 1:
            logging.critical('Multiple runs selected - using the first')
        selected = selected.first()
        return constants.Status(selected.status)

    def set_run_statuses(self, run_id_list, status):
        """Set the specified 'status' (enum) for all runs in the list run_id_list

        Parameters
        ----------
        run_id_list: list of int
            a list of run ids
        status: enum(Status)
            The new status all listed runs should now have
        """
        self.session.query(RunTable).filter(
            RunTable.id.in_(run_id_list)).update(
                {RunTable.status: status}, synchronize_session='fetch')
        self.session.commit()

    def campaigns(self):
        """Get list of campaigns for which information is stored in the
        database.

        Returns
        -------
        list
            Campaign names.
        """

        return [c.name for c in self.session.query(CampaignTable).all()]

    def _get_campaign_info(self, campaign_name=None):
        """Retrieves Campaign info based on name.

        Parameters
        ----------
        campaign_name: str
            Name of campaign to select.

        Returns
        -------
            SQLAlchemy query for campaign with this name.
        """
        assert(isinstance(campaign_name, str) or campaign_name is None)
        query = self.session.query(CampaignTable)
        if campaign_name is None:
            campaign_info = query
        else:
            campaign_info = query.filter_by(name=campaign_name).all()
        if campaign_name is not None:
            if len(campaign_info) > 1:
                logger.warning(
                    'More than one campaign selected - using first one.')
            elif len(campaign_info) == 0:
                message = 'No campaign available.'
                logger.critical(message)
                raise RuntimeError(message)
            return campaign_info[0]
        return campaign_info.first()

    def get_campaign_id(self, name):
        """Return the (database) id corresponding to the campaign with name 'name'.

        Parameters
        ----------
        name: str
            Name of the campaign.

        Returns
        -------
        int
            The id of the campaign with the specified name
        """

        selected = self.session.query(
            CampaignTable.name.label(name),
            CampaignTable.id).filter(CampaignTable.name == name).all()
        if len(selected) == 0:
            msg = f"No campaign with name {name} found in campaign database"
            logger.error(msg)
            raise RuntimeError(msg)
        if len(selected) > 1:
            msg = (
                f"More than one campaign with name {name} found in"
                f"campaign database. Database state is compromised."
            )
            logger.error(msg)
            raise RuntimeError(msg)
        # Return the database ID for the specified campaign
        return selected[0][1]

    def get_sampler_id(self, campaign_id):
        """Return the (database) id corresponding to the sampler currently set
        for the campaign with id 'campaign_id'

        Parameters
        ----------
        campaign_id: int
            ID of the campaign.

        Returns
        -------
        int
            The id of the sampler set for the specified campaign
        """
        sampler_id = self.session.query(CampaignTable).get(campaign_id).sampler
        return sampler_id

    def set_sampler(self, campaign_id, sampler_id):
        """Set specified campaign to be using specified sampler

        Parameters
        ----------
        campaign_id: int
            ID of the campaign.
        sampler_id: int
            ID of the sampler.
        """
        self.session.query(CampaignTable).get(campaign_id).sampler = sampler_id
        self.session.commit()

    def campaign_dir(self, campaign_name=None):
        """Get campaign directory for `campaign_name`.

        Parameters
        ----------
        campaign_name: str
            Name of campaign to select

        Returns
        -------
        str
            Path to campaign directory.
        """
        return self._get_campaign_info(campaign_name=campaign_name).campaign_dir

    def _select_runs(
            self,
            name=None,
            campaign=None,
            sampler=None,
            status=None,
            not_status=None,
            app_id=None):
        """Select all runs in the database which match the input criteria.

        Parameters
        ----------
        name: str
            Name of run to filter for.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string
        app_id: int or None
            App id to filter for.

        Returns
        -------
        sqlalchemy.orm.query.Query
            Selected runs from the database run table.
        """
        filter_options = {}
        if name:
            filter_options['run_name'] = name
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler
        if status:
            filter_options['status'] = status
        if app_id:
            filter_options['app'] = app_id

        # Note that for some databases this can be sped up with a yield_per(), but not all
        selected = self.session.query(RunTable).filter_by(
            **filter_options).filter(RunTable.status != not_status)

        return selected

    def run(self, name, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """Get the information for a specified run.

        Parameters
        ----------
        name: str
            Name of run to filter for.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string
        app_id: int or None
            App id to filter for.

        Returns
        -------
        dict
            Containing run information (run_name, params, status, sample,
            campaign, app)
        """
        selected = self._select_runs(
            name=name,
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)
        if selected.count() != 1:
            logging.warning('Multiple runs selected - using the first')
        selected = selected.first()
        return self._run_to_dict(selected)

    def runs(self, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """A generator to return all run information for selected `campaign` and `sampler`.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string
        app_id: int or None
            App id to filter for.

        Yields
        ------
        dict
            Information on each selected run (key = run_name, value = dict of
            run information fields.), one at a time.
        """
        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)
        for r in selected:
            yield r.id, self._run_to_dict(r)

    def run_ids(self, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """A generator to return all run IDs for selected `campaign` and `sampler`.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string
        app_id: int or None
            App id to filter for.

        Yields
        ------
        str
            run ID for each selected run, one at a time.
        """
        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)
        for r in selected:
            yield r.run_name

    def get_num_runs(self, campaign=None, sampler=None, status=None, not_status=None):
        """Returns the number of runs matching the filtering criteria.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        int
            The number of runs in the database matching the filtering criteria

        """
        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status)
        return selected.count()

    def runs_dir(self, campaign_name=None):
        """Get the directory used to store run information for `campaign_name`.

        Parameters
        ----------
        campaign_name: str
            Name of the selected campaign.

        Returns
        -------
        str
            Path containing run outputs.
        """
        return self._get_campaign_info(campaign_name=campaign_name).runs_dir

    def store_result(self, run_id, result, change_status=True):
        """Stores results of a simulation inside the RunTable given a run id.

        Parameters
        ----------
        run_id: int
            The id of a run to store the results in. This will be the run with which these
            results are associated with. Namely the run that has the inputs used to generate
            these results.
        result: dict
            Results in dictionary form. This is the same format as used by the `Decoder`.
        change_status: bool
            If set to False will not update the runs' status to COLLATED. This is sometimes
            useful in scenarios where you want several apps to work on the same runs.
        """
        self.commit_counter += 1

        def convert_nonserializable(obj):
            if isinstance(obj, np.int64):
                return int(obj)
            raise TypeError('Unknown type:', type(obj))
        result_ = result['result']
        result.pop('result')
        result.pop('run_info')
        if change_status:
            self.session.query(RunTable).\
                filter(RunTable.id == run_id).\
                update({'result': json.dumps(result_, default=convert_nonserializable),
                        'status': constants.Status.COLLATED,
                        'run_dir': result['rundir']})
        else:
            self.session.query(RunTable).\
                filter(RunTable.id == run_id).\
                update({'result': json.dumps(result_, default=convert_nonserializable),
                        'run_dir': result['rundir']})
        if self.commit_counter % COMMIT_RATE == 0:
            self.session.commit()

    def store_results(self, app_name, results):
        """Stores the results from a given run in the database.

        Parameters
        ----------
        run_name: str
            name of the run
        results: dict
            dictionary with the results (from the decoder)
        """
        try:
            app_id = self.session.query(AppTable).filter(AppTable.name == app_name).all()[0].id
        except IndexError:
            raise RuntimeError("app with the name {} not found".format(app_name))
        commit_counter = 0
        for run_id, result in results:
            try:
                self.session.query(RunTable).\
                    filter(RunTable.id == run_id, RunTable.app == app_id).\
                    update({'result': json.dumps(result), 'status': constants.Status.COLLATED})
                commit_counter += 1
                if commit_counter % COMMIT_RATE == 0:
                    self.session.commit()
            except IndexError:
                raise RuntimeError("no runs with name {} found".format(run_id))
        self.session.commit()

    def get_results(self, app_name, sampler_id, status=constants.Status.COLLATED, iteration=-1):
        """Returns the results as a pandas DataFrame.

        Parameters
        ----------
        app_name: str
            Name of the app to return data for.
        sampler_id: int
            ID of the sampler.
        status: STATUS
            Run status to filter for.
        iteration: int
            If a positive integer will return the results for a given iteration only.

        Returns
        -------
        DataFrame
            Will construct a `DataFrame` from the decoder output dictionaries.
        """
        try:
            app_id = self.session.query(AppTable).filter(AppTable.name == app_name).all()[0].id
        except IndexError:
            raise RuntimeError("app with the name {} not found".format(app_name))
        pd_result = {}
        query = self.session.query(RunTable).\
            filter(RunTable.app == app_id).\
            filter(RunTable.sampler == sampler_id).\
            filter(RunTable.status == status)
        # if only a specific iteration is requested filter it out
        if iteration >= 0:
            query = query.filter(RunTable.iteration == iteration)
        for row in query:
            params = {'run_id': row.id}
            params['iteration'] = row.iteration
            params = {**params, **json.loads(row.params)}
            result = json.loads(row.result)
            pd_dict = {**params, **result}
            for key in pd_dict.keys():
                if not isinstance(pd_dict[key], list):
                    try:
                        pd_result[(key, 0)].append(pd_dict[key])
                    except KeyError:
                        pd_result[(key, 0)] = [pd_dict[key]]
                else:
                    for i, elt in enumerate(pd_dict[key]):
                        try:
                            pd_result[(key, i)].append(pd_dict[key][i])
                        except KeyError:
                            pd_result[(key, i)] = [pd_dict[key][i]]
        try:
            return pd.DataFrame(pd_result)
        except ValueError:
            raise RuntimeError(
                'the results received from the database seem to be malformed - commonly because a vector quantity of interest changes dimensionality')

    def relocate(self, new_path, campaign_name):
        """Update all runs in the db with the new campaign path.

        Parameters
        ----------
        new_path: str
            new runs directory
        campaign_name: str
            name of the campaign
        """
        campaign_id = self.get_campaign_id(campaign_name)
        campaign_info = self.session.query(CampaignTable).\
            filter(CampaignTable.id == campaign_id).first()
        path, runs_dir = os.path.split(campaign_info.runs_dir)
        self.session.query(CampaignTable).\
            filter(CampaignTable.id == campaign_id).\
            update({'campaign_dir': str(new_path),
                    'runs_dir': str(os.path.join(new_path, runs_dir))})
        self.session.commit()

    def dump(self):
        """Dump the database as JSON for debugging purposes.

        Returns
        -------
        dict
            A database dump in JSON format.
        """
        meta = MetaData()
        meta.reflect(bind=self.engine)
        result = {}
        for table in meta.sorted_tables:
            result[table.name] = [dict(row) for row in self.engine.execute(table.select())]
        return json.dumps(result)



"""Will create a Gaussian Process surrogate of your model. For 
the sampler you can use the random sampler or the quasi-random
sampler. Don't forget to set the analysis class to GaussianProcessSurrogate
as is shown in the example below.

This uses the Gaussian Process model from sklearn.

Examples
--------
>>> campaign = uq.Campaign(name='surrogate')
>>> sampler = uq.sampling.RandomSampler(
    vary = {"Pe": cp.Uniform(100.0, 200.0), "f": cp.Uniform(0.95, 1.05)}
    max_num=100, analysis_class=uq.analysis.GaussianProcessSurrogate)
>>> campaign.add_app(name="sc", params=params, actions=actions)
>>> campaign.set_sampler(sampler)
>>> campaign.execute().collate()
>>> results = campaign.analyse(qoi_cols=output_columns)
>>> surrogate = results.surrogate()
>>> surrogate({'Pe' : 110.0, 'f': 1.0})
"""

from .base import BaseAnalysisElement
from sklearn.gaussian_process import GaussianProcessRegressor
from .results import AnalysisResults
import numpy as np

class GaussianProcessSurrogateResults(AnalysisResults):
    """Gaussian process surrogate results class. You would never
    create this manually in normal use. It is meant to be returned as the
    result of GaussianProcessSurrogate analyse method.

    Parameters
    ----------
    gps: list
        This will be one GP model for each coordinate of a vector QoI.
    parameters: list
        A list of input parameter names.
    qoi: str
        Output variable name.
    """
    def __init__(self, gp, parameters, qoi):
        self.gp = gp
        self.parameters = parameters
        self.qoi = qoi

    def surrogate(self):
        """Returns the GP surrogate model as a Python function.

        Returns
        -------
        function
            Returns a function that takes a dictionary and returns a dictionary.
            These dictionaries use the same format as Encoder and Decoder used
            to construct the surrogate.
        """
        def surrogate_fn(inputs):
            values = np.array([[inputs[key] for key in self.parameters]])
            return {self.qoi[0]: [x for x in self.gp.predict(values)[0]]}
        return surrogate_fn

    def get_params(self):
        return self.gp.kernel_.get_params()


class GaussianProcessSurrogate(BaseAnalysisElement):

    def __init__(self, sampler, qoi_cols, **kwargs):
        """An analysis class that can construct a Gaussian Process surrogate
        of your model. Based on the sklearn GaussianProgressRegressor class.

        Parameters
        ----------
        sampler : Sampler
            `Sampler` that was used to generate samples to train this surrogate model.
        qoi_cols : list
            Corresponding target values (can be vectors).
        """
        self.sampler = sampler
        self.attr_cols = list(sampler.vary.get_keys())
        self.target_cols = qoi_cols
        self.kwargs = kwargs

    def analyse(self, data_frame=None):
        """Construct a Gaussian Process surrogate based on data in `data_frame`.

        Parameters
        ----------
        data_frame : pandas.DataFrame
            Data which you want to use to fit the Gaussian Process to.
        kwargs : keyword arguments
            These arguments will be passed to sklearn's GaussianProcessRegressor.
            For details on what this could be, please see

        Returns
        -------
        easyvvuq.analysis.gp.GaussianProcessSurrogateResults
           `GaussianProcessSurrogateResults` instance. Used to interact with the surrogate
           model and to possibly access other functionality provided by the fitted model.
        """
        x = data_frame[self.attr_cols].values #lgtm [py/hash-unhashable-value]
        y = data_frame[self.target_cols].values #lgtm [py/hash-unhashable-value]
        gp = GaussianProcessRegressor(**self.kwargs)
        gp = gp.fit(x, y)
        return GaussianProcessSurrogateResults(gp, self.attr_cols, self.target_cols)



"""EasyVVUQ Campaign

This module contains the Campaign class that is used to coordinate all
EasyVVUQ workflows.
"""
import os
import logging
import tempfile
import easyvvuq
from concurrent.futures import ProcessPoolExecutor
from easyvvuq.constants import default_campaign_prefix, Status
from easyvvuq.data_structs import RunInfo, CampaignInfo, AppInfo
from easyvvuq.sampling import BaseSamplingElement
from easyvvuq.actions import ActionPool
import easyvvuq.db.sql as db

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.getLogger(__name__)


class Campaign:
    """Campaigns organise the dataflow in EasyVVUQ workflows.

    The Campaign functions as as state machine for the VVUQ workflows. It uses a
    database (CampaignDB) to store information on both the target application
    and the VVUQ algorithms being employed. It also collects data from the simulations
    and can be used to store and resume your state.

    Notes
    -----

    Multiple campaigns can be combined in a CampaignDB. Hence the particular
    campaign we are currently working on will be specified using `campaign_id`.

    Parameters
    ----------
    name: str
        Name of the Campaign. Freely chosen, serves as a human-readable way of distinguishing
        between several campaigns in the same database.
    params: dict, optional
        Description of the parameters to associated with the application. Will be used to create
        an app when creating the campaign. It is also possible to add apps manually using `add_app`
        method of the Campaign class. But this can be a useful shorthand when working with single
        app campaigns. To use this functionality both `params` and `actions` has to be specified.
        The name of this app will be the same as the name of the Campaign.
    actions: Actions, optional
        Actions object associated with an application. See description of the `params` parameter
        for more details.
    db_location: str, optional
        Location of the underlying campaign database - either a path or
        acceptable URI for SQLAlchemy.
    work_dir: str, optional, default='./'
        Path to working directory - used to store campaign directory.
    state_file: str, optional
        Path to serialised state - used to initialise the Campaign.
    change_to_state: bool, optional, default=False
        Should we change to the directory containing any specified `state_file`
        in order to make relative paths work.
    verify_all_runs: bool, optional, default=True
        Check all new runs being added for unrecognised params (not defined for the currently set
        app), values lying within defined physical range, type checking etc. This should normally
        always be set to True, but in cases where the performance is too degraded, the checks can
        be disabled by setting to False.

    Attributes
    ----------
    campaign_name : str or None
        Name for the campaign/workflow.
    _campaign_dir: str or None
        Path to the directory campaign uses for local storage (runs inputs etc)
    db_location : str or None
        Location of the underlying campaign database - either a path or
        acceptable URI for SQLAlchemy.
    _log: list
        The log of all elements that have been applied, with information about
        their application
    campaign_id : int
        ID number for the current campaign in the db.CampaignDB.
    campaign_db: easyvvuq.db.Basedb.CampaignDB
        A campaign database object
    last_analysis:
        The result of the most recent analysis carried out on this campaign
    _active_app: dict
        Info about currently set app
    _active_app_name: str
        Name of currently set app
    _active_sampler_id: int
        The database id of the currently set Sampler object

    Examples
    --------
    A typical instantiation might look like this.

    >>> params = {
            "S0": {"type": "float", "default": 997},
            "I0": {"type": "float", "default": 3},
            "beta": {"type": "float", "default": 0.2},
            "gamma": {"type": "float", "default": 0.04, "min": 0.0, "max": 1.0},
            "iterations": {"type": "integer", "default": 100},
            "outfile": {"type": "string", "default": "output.csv"}
        }
    >>> encoder = uq.encoders.GenericEncoder(template_fname='sir.template', delimiter='$', target_filename='input.json')
    >>> decoder = uq.decoders.SimpleCSV(target_filename='output.csv', output_columns=['I'])
    >>> actions = uq.actions.local_execute(encoder, os.path.abspath('sir') + ' input.json', decoder)
    >>> campaign = uq.Campaign(name='sir', params=params, actions=actions)

    A simplified one (without an app) might look simply like this.

    >>> campaign = Campaign('simple')

    An app then can be added.

    >>> campaign.add_app('simple_app', params=params, actions=actions)
    """

    def __init__(
            self,
            name,
            params=None,
            actions=None,
            db_location=None,
            work_dir="./",
            state_file=None,
            change_to_state=False,
            verify_all_runs=True
    ):

        self.work_dir = os.path.realpath(os.path.expanduser(work_dir))
        self.verify_all_runs = verify_all_runs

        self.campaign_name = name
        self._campaign_dir = None

        if db_location is None:
            self._campaign_dir = tempfile.mkdtemp(prefix=name, dir=work_dir)
            self.db_location = "sqlite:///" + self._campaign_dir + "/campaign.db"
        else:
            self.db_location = db_location

        self.campaign_id = None
        self.campaign_db = None

        self.last_analysis = None

        self._active_app = None
        self._active_app_name = None
        self._active_app_actions = None

        self._active_sampler = None
        self._active_sampler_id = None

        self.init_db(name, self.work_dir)
        self._state_dir = None

        # here we assume that the user wants to add an app
        if (params is not None) and (actions is not None):
            self.add_app(name=name, params=params, actions=actions)

    @property
    def campaign_dir(self):
        """Get the path in which to load/save files related to the campaign.

        Returns
        -------
        str
            Path to the campaign directory - given as a subdirectory of the
            working directory.
        """

        return os.path.join(self.work_dir, self._campaign_dir)

    def init_db(self, name, work_dir='.'):
        """Initialize the connection with the database and either resume or create the campaign.

        Parameters
        ----------
        name: str
            Name of the campaign.
        work_dir: str
            Work directory, defaults to cwd.
        """
        self.campaign_db = db.CampaignDB(location=self.db_location)
        if self.campaign_db.campaign_exists(name):
            self.campaign_id = self.campaign_db.get_campaign_id(name)
            self._active_app_name = self.campaign_db.get_active_app()[0].name
            self.campaign_name = name
            self._campaign_dir = self.campaign_db.campaign_dir(name)
            if not os.path.exists(self._campaign_dir):
                message = (f"Campaign directory ({self.campaign_dir}) does not exist.")
                raise RuntimeError(message)
            self._active_sampler_id = self.campaign_db.get_sampler_id(self.campaign_id)
            self._active_sampler = self.campaign_db.resurrect_sampler(self._active_sampler_id)
            self.set_app(self._active_app_name)
            self.campaign_db.resume_campaign(name)
        else:
            if self._campaign_dir is None:
                self._campaign_dir = tempfile.mkdtemp(prefix=name, dir=work_dir)
            info = CampaignInfo(
                name=name,
                campaign_dir_prefix=default_campaign_prefix,
                easyvvuq_version=easyvvuq.__version__,
                campaign_dir=self._campaign_dir)
            self.campaign_db.create_campaign(info)
            self.campaign_name = name
            self.campaign_id = self.campaign_db.get_campaign_id(self.campaign_name)

    def add_app(self, name=None, params=None, actions=None, set_active=True):
        """Add an application to the CampaignDB.

        Parameters
        ----------
        name : str
            Name of the application.
        params : dict
            Description of the parameters to associate with the application.
        actions : Actions
            An instance of Actions containing actions to be executed
        set_active: bool
            Should the added app be set to be the currently active app?
        """
        # Verify input parameters dict
        paramsspec = easyvvuq.ParamsSpecification(params, appname=name)
        # validate application input
        app = AppInfo(
            name=name,
            paramsspec=paramsspec,
            actions=actions,
        )
        self.campaign_db.add_app(app)
        if set_active:
            self.set_app(app.name)

    def set_app(self, app_name):
        """Set active app for the campaign.

        Application information is retrieved from `self.campaign_db`.

        Parameters
        ----------
        app_name: str
            Name of selected app, if `None` given then first app will be
            selected.
        """
        self._active_app_name = app_name
        self._active_app = self.campaign_db.app(name=app_name)
        self.campaign_db.set_active_app(app_name)
        # Resurrect the app encoder, decoder and collation elements
        self._active_app_actions = self.campaign_db.resurrect_app(app_name)

    def replace_actions(self, app_name, actions):
        """Replace actions for an app with a given name.

        Parameters
        ----------
        app_name: str
            Name of the app.
        actions: Actions
            `Actions` instance, will replace the current `Actions` of an app.
        """
        self.campaign_db.replace_actions(app_name, actions)
        self._active_app_actions = actions

    def set_sampler(self, sampler, update=False):
        """Set active sampler.

        Parameters
        ----------
        sampler : Sampler
            Sampler that will be used to create runs for the current campaign.
        update : bool
            If set to True it will not add the sampler to the database, just change
            it as the active sampler.
        """
        self._active_sampler = sampler
        if not update:
            self._active_sampler_id = self.campaign_db.add_sampler(sampler)
            sampler.sampler_id = self._active_sampler_id
        self._active_sampler_id = self._active_sampler.sampler_id
        self.campaign_db.set_sampler(self.campaign_id, self._active_sampler.sampler_id)

    def add_runs(self, runs, mark_invalid=False):
        """Add runs to the database.

        Parameters
        ----------
        runs : list of dicts
            Each dict defines the value of each model parameter listed in
            self.params_info for a run to be added to self.runs
        mark_invalid : bool
            Will mark runs that fail verification as invalid (but will not raise an exception)
        """
        if self._active_app is None:
            msg = ("No app is currently set for this campaign. "
                   "Use set_app('name_of_app').")
            logging.error(msg)
            raise Exception(msg)
        app_default_params = self._active_app["params"]
        run_info_list = []
        for new_run in runs:
            if new_run is None:
                msg = ("add_run() was passed new_run of type None. Bad sampler?")
                logging.error(msg)
                raise Exception(msg)
            # Verify and complete run with missing/default param values
            status = Status.NEW
            try:
                new_run = app_default_params.process_run(new_run, verify=self.verify_all_runs)
            except RuntimeError:
                if mark_invalid:
                    new_run = app_default_params.process_run(new_run, verify=False)
                    status = Status.INVALID
                else:
                    raise
            # Add to run queue
            run_info = RunInfo(app=self._active_app['id'],
                               params=new_run,
                               sample=self._active_sampler_id,
                               campaign=self.campaign_id,
                               status=status)
            run_info_list.append(run_info)
        self.campaign_db.add_runs(run_info_list, iteration=self._active_sampler.iteration)

    def draw_samples(self, num_samples=0, mark_invalid=False):
        """Draws `num_samples` sets of parameters from the currently set
        sampler, resulting in `num_samples` new runs added to the
        runs list. If `num_samples` is 0 (its default value) then
        this method draws ALL samples from the sampler, until exhaustion (this
        will fail if the sampler is not finite).

        Parameters
        ----------
        num_samples : int
            Number of samples to draw from the active sampling element.
            By default is 0 (draw ALL samples)
        mark_invalid : bool
            If True will mark runs that go outside valid parameter range as INVALID.
            This is useful for MCMC style methods where you want those runs to evaluate
            to low probabilities.
        """
        # Make sure `num_samples` is not 0 for an infinite generator
        # (this would add runs forever...)
        if not self._active_sampler.is_finite() and num_samples <= 0:
            msg = (f"Sampling_element '{self._active_sampler.element_name()}' "
                   f"is an infinite generator, therefore a finite number of "
                   f"draws (n > 0) must be specified.")
            raise RuntimeError(msg)
        num_added = 0
        new_runs = []
        for new_run in self._active_sampler:
            new_runs.append(new_run)
            num_added += 1
            if num_samples != 0 and num_added >= num_samples:
                break
        self.add_runs(new_runs, mark_invalid)
        # Write sampler's new state to database
        self.campaign_db.update_sampler(self._active_sampler_id, self._active_sampler)
        return new_runs

    def list_runs(self, sampler=None, campaign=None, app_id=None, status=None):
        """Get list of runs in the CampaignDB.

        Parameters
        ----------
        sampler: int
            Sampler id to filter for.
        campaign: int
            Campaign id to filter for.
        app_id: int
            App id to filter for.
        status: Status
            Status to filter for.

        Returns
        -------
        list of runs
        """
        return list(self.campaign_db.runs(
            sampler=sampler, campaign=campaign, app_id=app_id, status=status))

    def get_campaign_runs_dir(self):
        """Get the runs directory from the CampaignDB.

        Returns
        -------
        str
            Path in which the runs information will be written.
        """
        return self.campaign_db.runs_dir(self.campaign_name)

    def relocate(self, campaign_dir):
        """Relocate the campaign by specifying a new path where campaign is located.

        Parameters
        ----------
        new_path: str
            new runs directory
        """
        if not os.path.exists(campaign_dir):
            raise RuntimeError("specified directory does not exist: {}".format(campaign_dir))
        self.campaign_db.relocate(campaign_dir, self.campaign_name)

    def execute(self, nsamples=0, pool=None, mark_invalid=False, sequential=False):
        """This will draw samples and execute the Actions on those samples.

        Parameters
        ----------
        nsamples: int
            Number of samples to draw. For infinite samplers or when you want to process
            samples in batches.
        pool: Executor
            A pool object to be used when processing runs (e.g. instance of `ThreadPoolExecutor` or
            `ProcessPoolExecutor`).
        mark_invalid: bool
            Mark runs that go outside the specified input parameter range as INVALID.
        sequential: bool
            Whether to process samples sequentially (sometimes more efficient or you might
            want to avoid the concurrent module for some reason).
        """
        self.draw_samples(nsamples, mark_invalid=mark_invalid)
        action_pool = self.apply_for_each_sample(
            self._active_app_actions, sequential=sequential)
        return action_pool.start(pool=pool)

    def apply_for_each_sample(self, actions, status=Status.NEW, sequential=False):
        """For each run in this Campaign's run list, apply the specified action
        (an object of type Action).

        Parameters
        ----------
        actions: Actions
            Actions to be applied to each relevant run in the database.
        status: Status
            Will apply the Actions only to those runs whose status is as specified.
        sequential: bool
            Whether to process samples sequentially (sometimes more efficient or you might
            want to avoid the concurrent module for some reason).

        Returns
        -------
        ActionPool
            An object containing ActionStatus instances to track action execution.
        """
        # Loop through all runs in this campaign with status ENCODED, and
        # run the specified action on each run's dir
        def inits():
            for run_id, run_data in self.campaign_db.runs(
                    status=status, app_id=self._active_app['id']):
                previous = {}
                previous['run_id'] = run_id
                previous['campaign_dir'] = self._campaign_dir
                previous['rundir'] = run_data['run_dir']
                previous['run_info'] = run_data
                previous['result'] = {}
                previous['collated'] = False
                yield previous
        return ActionPool(self, actions, inits=inits(), sequential=sequential)

    def iterate(self, nsamples=0, pool=None, mark_invalid=False, sequential=False):
        """This is the equivalent of `execute` for methods that rely on the output of the
        previous sampling stage (designed for MCMC, should work for others).

        Parameters
        ----------
        nsamples : int
            Number of samples to draw (during a single iteration).
        pool : Executor
            An Executor instance. For example ThreadPoolExecutor or a Dask Client. Defaults
            to the ThreadPoolExecutor.
        mark_invalid : bool
            Mark runs that go outside the specified input parameter range as INVALID.
        sequential: bool
            Will execute the `Actions` associated with runs sequentially. Might be more
            efficient in some situations.

        Yields
        ------
        ActionPool
            An object containing Futures instances to track action execution.
        """
        while True:
            self.draw_samples(nsamples, mark_invalid=mark_invalid)
            action_pool = self.apply_for_each_sample(
                self._active_app_actions, sequential=sequential)
            yield action_pool.start(pool=pool)
            result = self.get_collation_result(last_iteration=True)
            invalid = self.get_invalid_runs(last_iteration=True)
            ignored_runs = self._active_sampler.update(result, invalid)
            for run_id in ignored_runs:
                self.campaign_db.session.query(db.RunTable).\
                    filter(db.RunTable.id == int(run_id)).\
                    update({'status': easyvvuq.constants.Status.IGNORED})
            self.campaign_db.session.commit()

    def recollate(self):
        """Clears the current collation table, changes all COLLATED status runs
           back to ENCODED, then runs collate() again
        """
        collated_run_ids = list(self.campaign_db.run_ids(status=Status.COLLATED))
        self.campaign_db.set_run_statuses(collated_run_ids, Status.ENCODED)
        self.collate()

    def get_collation_result(self, last_iteration=False):
        """Return dataframe containing all collated results

        Parameters
        ----------
        last_iteration : bool
            Will only return the result of the last iteration.

        Returns
        -------
        DataFrame
            A DataFrame with the simulation results along with the inputs
            used to produce them.
        """
        if last_iteration:
            iteration = self._active_sampler.iteration - 1
        else:
            iteration = -1
        return self.campaign_db.get_results(self._active_app['name'], self._active_sampler_id,
                                            status=easyvvuq.constants.Status.COLLATED, iteration=iteration)

    def get_invalid_runs(self, last_iteration=False):
        """Return dataframe containing all results marked as INVALID.

        Parameters
        ----------
        last_iteration : bool
            Will only return the result of the last iteration.

        Returns
        -------
        DataFrame
            A DataFrame with the results form simulations that were marked as INVALID.
            These will usually be the ones that went outside the specified parameter ranges.
            These still have to be accounted for in some way by some methods (e.g. MCMC).
        """
        if last_iteration:
            iteration = self._active_sampler.iteration - 1
        else:
            iteration = -1
        return self.campaign_db.get_results(self._active_app['name'], self._active_sampler_id,
                                            status=easyvvuq.constants.Status.INVALID, iteration=iteration)

    def apply_analysis(self, analysis):
        """Run the `analysis` element on the output of the last run collation.

        Parameters
        ----------
        analysis : Analysis
            Element that performs a VVUQ analysis on a dataframe summary of
            run outputs.
        """
        # Apply analysis element to most recent collation result
        self.last_analysis = analysis.analyse(data_frame=self.get_collation_result())

    def analyse(self, **kwargs):
        """If available will call an appropriate analysis class on the collation result.

        Parameters
        ----------
        **kwargs : dict
            Argument to the analysis class constructor (after sampler).

        Returns
        -------
        AnalysisResults
            An object representing analysis results. Can be used to interact with those results
            in some way. Plot, retrieve surrogate models and so on.
            See `easyvvuq.analysis.AnalysisResults` for further information.
        """
        collation_result = self.get_collation_result()
        try:
            analysis = self._active_sampler.analysis_class(sampler=self._active_sampler, **kwargs)
            return analysis.analyse(collation_result)
        except NotImplementedError:
            raise RuntimeError("This sampler does not have a corresponding analysis class")

    def get_last_analysis(self):
        """Return the output of the most recently run analysis element.
        """
        if self.last_analysis is None:
            logging.warning("No last analysis output available.")
        return self.last_analysis

    def __str__(self):
        """Returns formatted summary of the current Campaign state.
        Enables class to work with standard print() method
        """
        return (f"db_location = {self.db_location}\n"
                f"active_sampler_id = {self._active_sampler_id}\n"
                f"campaign_name = {self.campaign_name}\n"
                f"campaign_dir = {self.campaign_dir}\n"
                f"campaign_id = {self.campaign_id}\n")

    def get_active_sampler(self):
        """Return the active sampler element in use by this campaign.

        Returns
        -------
        The sampler currently in use
        """

        return self._active_sampler

    def ignore_runs(self, list_of_run_IDs):
        """Flags the specified runs to be IGNORED in future collation. Note that
        this does NOT remove previously collated results from the collation table.
        For that you must refresh the collation by running recollate().

        Parameters
        ----------
        list
            The list of run IDs for the runs that should be set to status IGNORED
        """
        self.campaign_db.set_run_statuses(list_of_run_IDs, Status.IGNORED)

    def rerun(self, list_of_run_IDs):
        """Sets the status of the specified runs to ENCODED, so that their results
        may be recollated later (presumably after extending, rerunning or otherwise
        modifying the data in the relevant run folder). Note that this method will
        NOT perform any execution - it simply flags the run in EasyVVUQ as being
        uncollated. Actual execution is (as usual) the job of the user or middleware.

        Parameters
        ----------
        list
            The list of run IDs for the runs that should be set to status ENCODED
        """

        for run_ID in list_of_run_IDs:
            status = self.campaign_db.get_run_status(run_ID)
            if status == Status.NEW:
                msg = (f"Cannot rerun {run_ID} as it has status NEW, and must"
                       f"be encoded before execution.")
                raise RuntimeError(msg)
        self.campaign_db.set_run_statuses(list_of_run_IDs, Status.ENCODED)

    def get_active_app(self):
        """Returns a dict of information regarding the application that is currently
        set for this campaign.
        """
        return self._active_app



import os
import easyvvuq as uq
import chaospy as cp
import pickle
import time
import numpy as np
import matplotlib.pylab as plt
from pprint import pprint
import glob
import subprocess
import pytest


@pytest.mark.skip(reason="fipy nonsense")
def test_action_replace():
    params = {
        "Qe_tot": {"type": "float", "min": 1.0e6, "max": 50.0e6, "default": 2e6},
        "H0": {"type": "float", "min": 0.00, "max": 1.0, "default": 0},
        "Hw": {"type": "float", "min": 0.01, "max": 100.0, "default": 0.1},
        "Te_bc": {"type": "float", "min": 10.0, "max": 1000.0, "default": 100},
        "chi": {"type": "float", "min": 0.01, "max": 100.0, "default": 1},
        "a0": {"type": "float", "min": 0.2, "max": 10.0, "default": 1},
        "R0": {"type": "float", "min": 0.5, "max": 20.0, "default": 3},
        "E0": {"type": "float", "min": 1.0, "max": 10.0, "default": 1.5},
        "b_pos": {"type": "float", "min": 0.95, "max": 0.99, "default": 0.98},
        "b_height": {"type": "float", "min": 3e19, "max": 10e19, "default": 6e19},
        "b_sol": {"type": "float", "min": 2e18, "max": 3e19, "default": 2e19},
        "b_width": {"type": "float", "min": 0.005, "max": 0.025, "default": 0.01},
        "b_slope": {"type": "float", "min": 0.0, "max": 0.05, "default": 0.01},
        "nr": {"type": "integer", "min": 10, "max": 1000, "default": 100},
        "dt": {"type": "float", "min": 1e-3, "max": 1e3, "default": 100},
        "out_file": {"type": "string", "default": "output.csv"}
    }

    encoder = uq.encoders.GenericEncoder(
        template_fname='tutorials/fusion.template',
        delimiter='$',
        target_filename='fusion_in.json')
    decoder = uq.decoders.SimpleCSV(
        target_filename="output.csv",
        output_columns=["te", "ne", "rho", "rho_norm"])
    execute = uq.actions.ExecuteLocal(
        'python3 %s/fusion_model.py fusion_in.json' % (os.getcwd()))
    actions = uq.actions.Actions(
        uq.actions.CreateRunDirectory('.'),
        uq.actions.Encode(encoder))
    vary = {
        "Qe_tot": cp.Uniform(1.8e6, 2.2e6),
        "Te_bc": cp.Uniform(80.0, 120.0)
    }
    campaign = uq.Campaign(name='fusion_pce.')
    campaign.add_app(name="fusion", params=params, actions=actions)
    campaign.set_sampler(uq.sampling.PCESampler(vary=vary, polynomial_order=2))
    campaign.execute().collate()
    input_json_files_PATH = [
        os.path.dirname(p)
        for p in glob.glob(
            campaign._campaign_dir + '/**/fusion_in.json',
            recursive=True
        )]
    curr_dir = os.getcwd()
    for run_dir in input_json_files_PATH:
        subprocess.call(
            ["python3 {}/tutorials/fusion_model.py  fusion_in.json".format(curr_dir)],
            cwd=run_dir,
            shell=True
        )
        print("executes fusion_model.py in directory {}".format(run_dir))
    actions = uq.actions.Actions(uq.actions.Decode(decoder))
    campaign.replace_actions("fusion", actions)
    campaign.execute().collate()
    results_df = campaign.get_collation_result()
    results = campaign.analyse(qoi_cols=["te", "ne", "rho", "rho_norm"])
    rho = results.describe('rho', 'mean')
    rho_norm = results.describe('rho_norm', 'mean')



"""
"""

from .base import BaseSamplingElement
import csv

__license__ = "LGPL"


class DataFrameSampler(BaseSamplingElement, sampler_name="csv_sampler"):

    def __init__(self, df, counter=0):
        """Takes a DataFrame and outputs it row by row.
        """
        self.data = df.to_dict(orient='records')
        self.counter = counter

    def is_finite(self):
        return True

    def n_samples(self):
        """Returns the number of samples in this sampler.
        Returns
        -------
        if the user specifies maximum number of samples than return that, otherwise - error
        """
        return len(self.data)

    def __next__(self):
        try:
            return self.data[self.counter]
        finally:
            if self.counter < self.n_samples():
                self.counter += 1
            else:
                raise StopIteration



"""Implements ActionPool - a thin wrapper around the Python Executor interface
that is meant to simplify the execution of actions and retrieval of results.
This object is instantiated by the Campaign. The user would never instantiate it
manually. The user does interact with it to track the progress of execution.
"""
import concurrent
from concurrent.futures import ThreadPoolExecutor
from dask.distributed import Client
from tqdm import tqdm
import copy

from . import QCGPJPool

__copyright__ = """

    Copyright 2020 Vytautas Jancauskas

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class ActionPool:
    """A class that handles the execution of Actions.

    Parameters
    ----------
    campaign: Campaign
        An instance of an EasyVVUQ campaign.
    actions: Actions
        An instance of `Actions` containing things to be done as part of the simulation.
    inits: iterable
        Initial inputs to be passed to each `Actions` representing a sample. Will usually contain
        dictionaries with the following information: {'run_id': ..., 'campaign_dir': ..., 
        'run_info': ...}.
    sequential: bool
        Will run the actions sequentially.
    """

    def __init__(self, campaign, actions, inits, sequential=False):
        self.campaign = campaign
        self.actions = actions
        self.inits = inits
        self.sequential = sequential
        self.futures = []
        self.results = []
        self._collate_callback = lambda previous: previous

    def start(self, pool=None):
        """Start the actions.

        Parameters
        ----------
        pool: An Executor instance (e.g. ThreadPoolExecutor)

        Returns
        -------
        ActionPool
            Starts execution and returns a reference to itself for tracking progress 
            and for collation.
        """
        if pool is None:
            pool = ThreadPoolExecutor()
        self.pool = pool
        for previous in self.inits:
            previous = copy.copy(previous)
            if self.sequential:
                result = self.actions.start(previous)
                self.results.append(result)
            else:
                future = self.pool.submit(self.actions.start, previous)
                self.futures.append(future)
        return self

    def progress(self):
        """Some basic stats about the action statuses status.

        Returns
        -------
        dict
            A dictionary with four keys - 'ready', 'active' and 'finished', 'failed'. 
            Values under "ready" correspond to `Actions` waiting for execution, "active"
            corresponds to the number of currently running tasks.
        """
        ready = 0
        running = 0
        done = 0
        failed = 0
        for future in self.futures:
            if future.running():
                running += 1
            elif future.done():
                if not future.result():
                    failed += 1
                else:
                    done += 1
            else:
                ready += 1
        return {'ready': ready, 'active': running, 'finished': done, 'failed': failed}

    def add_collate_callback(self, fn):
        """Adds a callback to be called after collation is done.
        
        Parameters
        ----------
        fn - A callable that takes previous as it's only input.
        """
        self._collate_callback = fn

    def collate(self, progress_bar=False):
        """A command that will block untill all Futures in the pool have finished.
        It will also store the results gather from `Actions` in the database.

        Parameters
        ----------
        progress_bar: bool
           Whether to show progress bar
        """
        if not progress_bar:
            def tqdm_(x, total=None): return x
        else:
            tqdm_ = tqdm
        if isinstance(self.pool, Client):
            self.results = self.pool.gather(self.futures)
        if self.sequential or isinstance(self.pool, Client):
            for result in tqdm_(self.results, total=len(self.results)):
                result = self._collate_callback(result)
                self.campaign.campaign_db.store_result(
                    result['run_id'], result, change_status=result['collated'])
        else:
            if isinstance(self.pool, QCGPJPool):
                as_completed_fn = self.pool.as_completed
                self.add_collate_callback(self.pool.convert_results)
            else:
                as_completed_fn = concurrent.futures.as_completed

            for future in tqdm_(as_completed_fn(self.futures), total=len(self.futures)):
                result = self._collate_callback(future.result())
                self.campaign.campaign_db.store_result(
                    result['run_id'], result, change_status=result['collated'])
        self.campaign.campaign_db.session.commit()



import pytest
import os
from easyvvuq.actions.execute_local import CreateRunDirectory


def test_create_run_directory(tmpdir):
    action = CreateRunDirectory(tmpdir)
    previous = {'campaign_dir': 'test', 'run_id': 123456789, 'run_info': {'id': 123456789}}
    previous = action.start(previous)
    assert(os.path.exists(
        os.path.join(
            tmpdir, 'test', 'runs', 'runs_100000000-200000000', 'runs_123000000-124000000',
            'runs_123450000-123460000', 'runs_123456700-123456800', 'run_123456789')))
    previous = {'campaign_dir': 'test', 'run_id': 0, 'run_info': {'id': 0}}
    previous = action.start(previous)
    assert(os.path.exists(
        os.path.join(
            tmpdir, 'test', 'runs', 'runs_0-100000000', 'runs_0-1000000', 'runs_0-10000',
            'runs_0-100')))
    previous = {'campaign_dir': 'test', 'run_id': 100, 'run_info': {'id': 100}}
    action = CreateRunDirectory(tmpdir, flatten=True)
    previous = action.start(previous)
    assert(os.path.exists(os.path.join(tmpdir, 'test', 'runs', 'run_100')))



"""Analysis element for the Markov Chain Monte Carlo (MCMC) method.
For more details on the method see the `easyvvuq.sampling.MonteCarloSampler` class.
The analysis part of Markov Chain Monte Carlo consists of approximating the distribution
from the results obtained by evaluating the samples.
"""
import pandas as pd
from .base import BaseAnalysisElement
from .results import AnalysisResults


class MCMCAnalysisResults(AnalysisResults):
    """The analysis results class for MCMC. You will not need to instantiate this
    class manually.

    Parameters
    ----------
    chains: dict
        A dictionary with pandas DataFrame that correspond to an MCMC chain each.
        A chain consists of points that MCMC has visited. From this a distribution
        of the input variables can be constructed by means of a simple histogram.
    """

    def __init__(self, chains):
        self.chains = chains

    def plot_hist(self, input_parameter, chain=None, skip=0, merge=True):
        """Will plot a histogram for a given input parameter.

        Parameters
        ----------
        input_parameter: str
            An input parameter name to draw the histogram for.
        chain: int, optional
            Index of a chain to be plotted.
        skip: int
            How many steps to skip (for getting rid of burn-in).
        merge: bool
            If set to True will use all chains to construct the histogram.
        """
        import matplotlib.pyplot as plt
        input_parameter = (input_parameter, 0)
        if merge:
            chain_keys = list(self.chains.keys())
            df = self.chains[chain_keys[0]][input_parameter].iloc[skip:]
            for chain in chain_keys[1:]:
                df.append(self.chains[chain][input_parameter].iloc[skip:])
            plt.hist(df, 20)
        else:
            plt.hist(self.chains[chain][input_parameter].iloc[skip:], 20)

    def plot_chains(self, input_parameter, chain=None):
        """Will plot the chains with the input parameter value in the y axis.
        
        Parameters
        ----------
        input_parameter: str
            Input parameter name.
        chain: int, optional
            The chain number of the chain to plot.
        """
        import matplotlib.pyplot as plt
        if chain is None:
            for chain in self.chains:
                plt.plot(self.chains[chain][(input_parameter, 0)])
        else:
            plt.plot(self.chains[chain][(input_parameter, 0)])


class MCMCAnalysis(BaseAnalysisElement):
    """The analysis part of the MCMC method in EasyVVUQ

    Parameters
    ----------
    sampler: MCMCSampler
       An instance of MCMCSampler used to generate MCMC samples.
    """
    def __init__(self, sampler):
        self.sampler = sampler

    def analyse(self, df):
        """Performs some pre-processing on the chains in order to be able to construct
        the histograms or other methods of distribution estimation.

        Parameters
        ----------
        df: DataFrame
            DataFrame with the results obtained by evaluating the samples generated by the
            MCMC sampler.
        """
        chains = dict([(chain_id, []) for chain_id in df[('chain_id', 0)].unique()])
        for chain in chains:
            chain_values = df[df[('chain_id', 0)] == chain]
            values = chain_values.groupby(('iteration', 0)).apply(lambda x: x.mean())
            indexes = values.index.values
            for a, b in zip(indexes[:-1], indexes[1:]):
                chains[chain] += [values.loc[a][self.sampler.inputs].to_dict()] * (b - a)
        for chain in chains:
            tmp = dict([(input_, []) for input_ in chains[chain][0]])
            for row in chains[chain]:
                for input_ in chains[chain][0]:
                    tmp[input_].append(row[input_])
            chains[chain] = pd.DataFrame(tmp)
        return MCMCAnalysisResults(chains)



import easyvvuq as uq
from easyvvuq.actions import CreateRunDirectory, Encode, ExecuteLocal, Decode, Actions
import chaospy as cp
import os
import logging
import pytest
import shutil
from easyvvuq.db.sql import CampaignTable

TEST_PATH = os.path.dirname(os.path.realpath(__file__))
LOGGER = logging.getLogger(__name__)


@pytest.fixture
def campaign(tmpdir):
    params = {
        "angle": {
            "type": "float",
            "min": 0.0,
            "max": 6.28,
            "default": 0.79},
        "air_resistance": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.2},
        "height": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 1.0},
        "time_step": {
            "type": "float",
            "min": 0.0001,
            "max": 1.0,
            "default": 0.01},
        "gravity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 9.8},
        "mass": {
            "type": "float",
            "min": 0.0001,
            "max": 1000.0,
            "default": 1.0},
        "velocity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 10.0}
    }
    encoder = uq.encoders.GenericEncoder(
        template_fname=f'{TEST_PATH}/cannonsim/test_input/cannonsim.template',
        target_filename='in.cannon')
    decoder = uq.decoders.SimpleCSV(
        target_filename='output.csv', output_columns=[
            'Dist', 'lastvx', 'lastvy'])
    execute = ExecuteLocal(f"{TEST_PATH}/cannonsim/bin/cannonsim in.cannon output.csv")
    campaign = uq.Campaign(name='test', work_dir=tmpdir)
    actions = Actions(CreateRunDirectory('/tmp'), Encode(encoder), execute, Decode(decoder))
    campaign.add_app(name='test', params=params, actions=actions)
    campaign.set_app('test')
    stats = uq.analysis.BasicStats(qoi_cols=['Dist', 'lastvx', 'lastvy'])
    # Make a random sampler
    vary = {
        "angle": cp.Uniform(0.0, 1.0),
        "height": cp.Uniform(2.0, 10.0),
        "velocity": cp.Normal(10.0, 1.0),
        "mass": cp.Uniform(1.0, 5.0)
    }
    sampler = uq.sampling.RandomSampler(vary=vary)
    campaign.set_sampler(sampler)
    campaign.execute(nsamples=100, sequential=True).collate()
    return campaign


def test_campaign_exists(tmp_path):
    campaign = uq.Campaign(name='test', work_dir=tmp_path)
    assert(campaign.campaign_db.campaign_exists('test'))
    assert(not campaign.campaign_db.campaign_exists('test2'))


def test_invalid_sampler(tmp_path):
    with pytest.raises(Exception):
        our_campaign = uq.Campaign(name='test', work_dir=tmp_path)
        our_campaign.set_sampler('not_a_sampler')


def test_premature_run_addition(tmp_path):
    with pytest.raises(Exception):
        our_campaign = uq.Campaign(name='test', work_dir=tmp_path)
        # Requires an active app to be set
        our_campaign.add_runs([])


def test_none_run_addition(tmp_path):
    with pytest.raises(Exception):
        our_campaign = uq.Campaign(name='test', work_dir=tmp_path)
        # Requires an active app to be set
        our_campaign.add_runs([None])


def test_premature_get_last_analysis(caplog, tmp_path):
    with caplog.at_level(logging.WARNING):
        our_campaign = uq.Campaign(name='test', work_dir=tmp_path)
        result = our_campaign.get_last_analysis()

    assert 'No last analysis output available.' in caplog.text
    assert result is None


def test_string(campaign):
    target_str = (
        f"db_location = {campaign.db_location}\n"
        f"active_sampler_id = 1\n"
        f"campaign_name = test\n"
        f"campaign_dir = {campaign.campaign_dir}\n"
        f"campaign_id = 1\n")
    assert str(campaign) == target_str


def test_get_active_sampler(tmp_path):
    camp = uq.Campaign(name='test', work_dir=tmp_path)
    vary = {
        "angle": cp.Uniform(0.0, 1.0),
        "height": cp.Uniform(2.0, 10.0),
        "velocity": cp.Normal(10.0, 1.0),
        "mass": cp.Uniform(1.0, 5.0)
    }
    sampler = uq.sampling.RandomSampler(vary=vary)
    camp.set_sampler(sampler)

    assert camp.get_active_sampler() == sampler


def test_get_active_app(campaign):
    assert campaign.get_active_app() == campaign._active_app



import easyvvuq as uq
import chaospy as cp
import os
import pytest
import shutil
from easyvvuq.actions import CreateRunDirectory, Encode, ExecuteLocal, Decode, Actions


def test_gp(tmp_path):
    campaign = uq.Campaign(name='test', work_dir=tmp_path)
    params = {
        "temp_init": {
            "type": "float",
            "min": 0.0,
            "max": 100.0,
            "default": 95.0},
        "kappa": {
            "type": "float",
            "min": 0.0,
            "max": 0.1,
            "default": 0.025},
        "t_env": {
            "type": "float",
            "min": 0.0,
            "max": 40.0,
            "default": 15.0},
        "out_file": {
            "type": "string",
            "default": "output.csv"}}
    output_filename = params["out_file"]["default"]
    output_columns = ["te"]
    # Create an encoder and decoder for PCE test app
    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cooling/cooling.template',
        delimiter='$',
        target_filename='cooling_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                    output_columns=output_columns)
    execute = ExecuteLocal("{} cooling_in.json".format(
        os.path.abspath("tests/cooling/cooling_model.py")))
    actions = Actions(CreateRunDirectory('/tmp'), Encode(encoder), execute, Decode(decoder))
    vary = {
        "kappa": cp.Uniform(0.025, 0.075),
        "t_env": cp.Uniform(15, 25)
    }
    sampler = uq.sampling.quasirandom.LHCSampler(vary=vary)

    campaign.add_app(name='test_app', params=params, actions=actions)
    campaign.set_app('test_app')
    campaign.set_sampler(sampler)
    campaign.execute(nsamples=100).collate()
    df = campaign.get_collation_result()
    analysis = uq.analysis.gp_analyse.GaussianProcessSurrogate(sampler, ['te'])
    result = analysis.analyse(df)



import base64
import json
import logging
from os import environ

import time

import dill
from typing import Tuple, Dict, Any

from concurrent.futures import Executor

from qcg.pilotjob.executor_api.qcgpj_executor import QCGPJExecutor
from qcg.pilotjob.executor_api.templates.qcgpj_template import QCGPJTemplate


logger = logging.getLogger(__name__)


class EasyVVUQBasicTemplate(QCGPJTemplate):
    """A basic template class for submission of QCG-PilotJob tasks that run on a single core

    The class can be used only for the most simple use-cases. For example it doesn't allow
    to specify resource requirements. Thus, for more advanced use-cases, it is recommended to provide custom
    implementation of QCGPJTemplate. For complete reference of QCG-PilotJob task's description parameters
    please look at https://qcg-pilotjob.readthedocs.io/en/latest/fileinterface.html#submit
    """
    @staticmethod
    def template() -> Tuple[str, Dict[str, Any]]:
        template = """
            {
                'name': '${name}',
                'execution': {
                    'exec': '${exec}',
                    'args': ${args},
                    'stdout': '${stdout}',
                    'stderr': '${stderr}',
                    'venv': '${venv}',
                    'model': '${model}',
                    'model_opts': ${model_opts}
                }
            }
             """

        defaults = {
            'args': [],
            'stdout': 'stdout',
            'stderr': 'stderr',
            'venv': '',
            'model': 'default',
            'model_opts': {}
        }

        return template, defaults


class EasyVVUQParallelTemplate(QCGPJTemplate):
    """A template class for submission of QCG-PilotJob tasks that run on exact number cores / nodes

    With this class it is possible to define basic resource requirements for tasks.
    For advanced use-cases, it is recommended to provide custom implementation of QCGPJTemplate.
    For complete reference of QCG-PilotJob task's description parameters
    please look at https://qcg-pilotjob.readthedocs.io/en/latest/fileinterface.html#submit
    """
    @staticmethod
    def template() -> Tuple[str, Dict[str, Any]]:
        template = """
            {
                'name': '${name}',
                'execution': {
                    'exec': '${exec}',
                    'args': ${args},
                    'stdout': '${stdout}',
                    'stderr': '${stderr}',
                    'venv': '${venv}',
                    'model': '${model}',
                    'model_opts': ${model_opts}
                },
                'resources': {
                    'numCores': {
                        'exact': ${numCores}
                    },
                    'numNodes': {
                        'exact': ${numNodes}
                    }
                }
            }
             """

        defaults = {
            'args': [],
            'stdout': 'stdout',
            'stderr': 'stderr',
            'venv': '',
            'model': 'default',
            'model_opts': {},
            'numCores': 1,
            'numNodes': 1
        }

        return template, defaults


class QCGPJPool(Executor):
    """A Pool that manages execution of actions with QCG-PilotJob.

    Parameters
    ----------
    qcgpj_executor: str
        An instance of QCGPJExecutor. If not provided, an instance of QCGPJExecutor
        with default settings will be created
    template: QCGPJTemplate
        An object which contains only a single method `template` that returns a tuple.
        The first element of a tuple should be a string representing a QCG-PilotJob task's description
        with placeholders (identifiers preceded by $ symbol)
        and the second a dictionary that assigns default values for selected placeholders.
        If not provided, a default EasyVVUQBasicTemplate will be used
    template_params: dict
        A dictionary that contains parameters that will be used to substitute placeholders
        defined in the template
    polling_interval: int
        An interval between queries to the QCG-PilotJob Manager service about state of the tasks, in seconds.
    """
    def __init__(self, qcgpj_executor=None, template=None, template_params=None, polling_interval=1):
        if qcgpj_executor is None:
            qcgpj_executor = QCGPJExecutor()
        if template is None:
            template = EasyVVUQBasicTemplate()

        self._qcgpj_executor = qcgpj_executor
        self._template = template
        self._template_params = template_params
        self._polling_interval = polling_interval
        self._campaign_dir = None

    def submit(self, fn, *args, **kwargs):
        """Submits a callable to be executed by QCG-PilotJob.

        Schedules the callable to be executed inside a QCG-PilotJob's task and returns
        a Future representing the execution of the callable.

        Returns
        -------
            QCGPJFuture representing the given call.
        """
        actions = fn.__self__
        actions.set_wrapper(QCGPJPool._wrapper)
        exec = 'python3 -m easyvvuq.actions.execute_qcgpj_task'

        pickled_actions = base64.b64encode(dill.dumps(actions)).decode('ascii')
        pickled_previous = base64.b64encode(dill.dumps(args[0])).decode('ascii')

        self._campaign_dir = args[0]['campaign_dir']

        return self._qcgpj_executor.submit(
            self._template.template,
            self._template_params,
            exec=exec,
            name=args[0]['run_id'],
            stdout=f"{self._campaign_dir}/stdout_{args[0]['run_id']}",
            stderr=f"{self._campaign_dir}/stderr_{args[0]['run_id']}",
            args=[pickled_actions, pickled_previous])

    @property
    def executor(self):
        """Returns current QCGPJExecutor instance.

        It gives you an access to QCG-PilotJob Manager instance, which in turn can be used to
        get information about the QCG-PilotJob execution environment.
        """
        return self._qcgpj_executor

    def convert_results(self, result_qcgpj):
        """Converts results generated by QCG-PilotJob task to EasyVVUQ-compatible form

        The method loads results data from a file where it was stored by QCG-PilotJob's task
        and then converts it to a dictionary which can be further processed by EasyVVUQ.

        Parameters
        ----------
        result_qcgpj: list or None
            A list of results returned by a QCG-PilotJob task (only the first element will be used),
            or None if the task hasn't finished with the status SUCCEED

        Returns
        -------
            A dictionary containing results
        """

        for key, value in result_qcgpj.items():
            if value != 'SUCCEED':
                logging.error(f"Task {key} finished with the status: {value}")
                raise RuntimeError(f"QCG-PilotJob task {key} finished with the status: {value}")
            with open(f'{self._campaign_dir}/.qcgpj_result_{key}', 'r') as f:
                previous = json.load(f)
                return previous

    def shutdown(self, **kwargs):
        """Clean-up the resources associated with the QCGPJPool.
        """
        return self._qcgpj_executor.shutdown()

    def as_completed(self, futures):
        """Checks for the status of features and yields those that are finished
        """

        pending = set(futures)
        finished = set()

        for f in futures:
            if f.done():
                finished.add(f)
        pending = pending - finished

        while finished:
            yield finished.pop()

        while pending:
            for f in pending:
                if f.done():
                    finished.add(f)
            pending = pending - finished

            while finished:
                yield finished.pop()

            time.sleep(self._polling_interval)

    @staticmethod
    def _wrapper(action, previous):
        """For the actions other than ExecuteQCGPJ ensures that the MPI code is invoked in a serial mode
        """
        if not isinstance(action, ExecuteQCGPJ):
            rank = 0
            if 'OMPI_COMM_WORLD_RANK' in environ:
                rank = environ.get('OMPI_COMM_WORLD_RANK')
            elif 'PMI_RANK' in environ:
                rank = environ.get('PMI_RANK')

            if rank != 0:
                # This is not an instance of ExecuteQCGPJ,We don't execute processes with ranks other than 0
                return

        return action.start(previous)


class ExecuteQCGPJ:
    """A utility decorator over action that marks the action to enable parallel execution by QCG-PilotJob

    Parameters
    ----------
    action: Action
        an action that will be decorated in order to enable parallel execution inside a QCG-PilotJob task.
    """
    def __init__(self, action):
        self._action = action

    def start(self, previous=None):
        return self._action.start(previous)

    def finished(self):
        return self._action.finished()

    def finalise(self):
        return self._action.finalise()

    def succeeded(self):
        return self._action.succeeded()



import base64
import json
import os
import sys

import dill

from . import Actions

if __name__ == "__main__":

    actions: Actions

    if len(sys.argv) != 3:
        sys.exit(
            "Usage: python3 encoded_actions_object encoded_previous_object"
        )

    if 'QCG_PM_EXEC_API_JOB_ID' not in os.environ:
        sys.exit("The required environment variable QCG_PM_STEP_ID not set")

    jobid = os.environ['QCG_PM_EXEC_API_JOB_ID']

    pickled_actions = sys.argv[1]
    pickled_actions_b64 = pickled_actions.encode('ascii')
    pickled_actions_b = base64.b64decode(pickled_actions_b64)
    actions = dill.loads(pickled_actions_b)

    pickled_previous = sys.argv[2]
    pickled_previous_b64 = pickled_previous.encode('ascii')
    pickled_previous_b = base64.b64decode(pickled_previous_b64)
    previous = dill.loads(pickled_previous_b)

    previous_next = actions.start(previous)

    with open(f'{previous_next["campaign_dir"]}/.qcgpj_result_{jobid}', 'w') as f:
        json.dump(previous_next, f)



"""Provides a simple action element for interacting with a SLURM job. This lets you
execute your simulation on a SLURM cluster.
"""

import logging
import re
import subprocess
import time
import os
import random

__license__ = "LGPL"

logger = logging.getLogger(__name__)


class ExecuteSLURM():
    """An Action to launch and track the execution of a SLURM job.

    Parameters
    ----------
    template_script: str
        Filename of a file containing the script template.
    variable: str
        A string to be replaced with the directory in which the job is meant to be executed.
        This is to be used to make sure that the simulation can find the correct input files and knows
        where to put output files.
    """

    def __init__(self, template_script, variable):
        with open(template_script, 'r') as fd:
            self.template = fd.read()
        self.script_name = template_script
        self.variable = variable

    def start(self, previous=None):
        """Start the SLURM job.

        Parameters
        ----------
        previous: dict
            A dictionary containing information provided by previously executed actions.
        """
        target_dir = previous['rundir']
        script_name = os.path.join(target_dir, os.path.basename(self.script_name))
        script, = self.template.replace(self.variable, target_dir),
        with open(script_name, 'w') as fd:
            fd.write(script)
        result = subprocess.run(
            ['sbatch', script_name],
            cwd=target_dir, check=True, capture_output=True)
        stdout = result.stdout.decode('utf-8')
        self.job_id = re.findall(r'\d+', stdout)[0]
        while True:
            result = subprocess.run(
                ['squeue', '-j', self.job_id],
                cwd=target_dir, check=True, capture_output=True)
            stdout = result.stdout.decode('utf-8')
            if self.job_id not in stdout:
                break
            time.sleep(random.randint(1, 600))
        return previous

    def finalise(self):
        """Performs clean-up if necessary. In this case it isn't. I think.
        """
        return None



import os
import easyvvuq as uq
import numpy as np
import chaospy as cp
import json
import matplotlib.pyplot as plt
import sys
from easyvvuq.actions import ExecutePython, Actions
from dask.distributed import Client


HOME = os.path.abspath(os.path.dirname(__file__))


def rosenbrock(inputs):
    x1 = float(inputs['x1'])
    x2 = float(inputs['x2'])
    y = (1.0 - x1) ** 2 + 100.0 * (x2 - x1 ** 2) ** 2
    return {'value': 300.0 - y}


def test_mcmc(tmp_path):
    campaign = uq.Campaign(name="mcmc", work_dir=tmp_path)
    params = {
        "x1": {"type": "float", "default": 0.0},
        "x2": {"type": "float", "default": 0.0},
        "out_file": {"type": "string", "default": "output.json"},
        "chain_id": {"type": "integer", "default": 0}
    }
    encoder = uq.encoders.GenericEncoder(template_fname=os.path.abspath(
        "tutorials/rosenbrock.template"), delimiter="$", target_filename="input.json")
    decoder = uq.decoders.JSONDecoder("output.json", ["value"])
    actions = Actions(ExecutePython(rosenbrock))
    campaign.add_app(name="mcmc", params=params, actions=actions)
    vary_init = {
        "x1": [-1.0, 0.0, 1.0, 0.5, 0.1],
        "x2": [1.0, 0.0, 0.5, 1.0, 0.2]
    }

    def q(x, b=1):
        return cp.J(cp.Normal(x['x1'], b), cp.Normal(x['x2'], b))
    np.random.seed(1969)
    sampler = uq.sampling.MCMCSampler(vary_init, q, 'value', 5)
    campaign.set_sampler(sampler)
    iterator = campaign.iterate(pool=Client())
    for _ in range(200):
        next(iterator).collate(progress_bar=True)
    df = campaign.get_collation_result()
    analysis = uq.analysis.MCMCAnalysis(sampler)
    result = analysis.analyse(df)
    result.plot_hist('x1')
    result.plot_hist('x2')
    result.plot_chains('x1')
    result.plot_chains('x2')
